//+------------------------------------------------------------------+
//|                                                      ProjectName |
//|                                      Copyright 2020, CompanyName |
//|                                       http://www.companyname.net |
//+------------------------------------------------------------------+
#property indicator_chart_window
#property indicator_buffers 8
#property indicator_plots 8

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
input double daysback = 2;//Amount of days worth of valid H/L
input double   margin = 0.25;//Set a tolerance margin to ignore minor price differences
input bool plotHighLow = true;     //Plot the valid highs and lows
input bool plotProtectedHighlow = false;  //Plot the protected highs and lows
input bool plotcurrentrange = true;//Plots the current trading range
input bool plotrangebreak = true;//Plots the level that will make a new zone
input bool plotfib = true;//Plots the fib levels
input bool plotVHLValidation = true;//Plots the level that if broken will confirm valid highs and lows
input double level1 = 30;  // Set fib level 1 in %
input double level2 = 50;  // Set fib level 2 in %
input double level3 = 70;  // Set fib level 3 in %
input bool useLevel1 = true;  // Toggle to use Level 1
input bool useLevel2 = true;  // Toggle to use Level 2
input bool useLevel3 = true;  // Toggle to use Level 3


double   HighsBuffer[];
double   LowsBuffer[];
double   LineBuffer[];
double   ProtectedHighsBuffer[];
double   ProtectedLowsBuffer[];
double   LineToFlipBullish[];
double   LineToFlipBearish[];
double   LineForNewZone[];

int      validHighIndex = -1;
int      validLowIndex = -1;
int      prevvalidHighindex =-1;
int      prevvalidLowindex =-1;
int      rangeValidHighIndex = -1;
int      rangeValidLowIndex = -1;
int      breakoutHighIndex=-1;
int      breakoutLowIndex=-1;
int      fibStartIndex=-1;
int      fibEndIndex = 1;

double   highestHigh=-1;
double   lowestLow=1E+18;
double   VLH=1E+18;
double   VLL=-1;
double   fibHighLevel=0;
double   fibLowLevel=0;
double   currentVC = EMPTY_VALUE;
double   currentLineToFlipBullish = EMPTY_VALUE;
double   currentLineToFlipBearish = EMPTY_VALUE;
double   currentLineForNewZone = EMPTY_VALUE;

bool     searchForHigh = true;
bool     gapup_search = true;
bool     gapdown_search = true;
bool     highfound = false;
bool     lowfound = false;
bool     range_determined = false;
bool     highbreached = false;
bool     lowbreached = false;
bool     waitingfornewrange = true;

struct FibZone
  {
   double            high;
   double            low;
   int               startIdx;
   int               endIdx;
   string            fibNames[3];  // Names of trend lines for each Fibonacci level
   bool              breached;       // Flag to indicate if a new zone has been confirmed
  };

FibZone fibZones[];


//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
int OnInit()
  {
   ArrayInitialize(HighsBuffer, EMPTY_VALUE);
   ArrayInitialize(LowsBuffer, EMPTY_VALUE);
   ArrayInitialize(LineBuffer, EMPTY_VALUE);
   ArrayInitialize(ProtectedHighsBuffer, EMPTY_VALUE);
   ArrayInitialize(ProtectedLowsBuffer, EMPTY_VALUE);
   ArrayInitialize(LineToFlipBullish, EMPTY_VALUE);
   ArrayInitialize(LineToFlipBearish, EMPTY_VALUE);
   ArrayInitialize(LineForNewZone, EMPTY_VALUE);

   SetIndexBuffer(0, HighsBuffer, INDICATOR_DATA);
   SetIndexBuffer(1, LowsBuffer, INDICATOR_DATA);
   SetIndexBuffer(2, LineBuffer, INDICATOR_DATA);
   SetIndexBuffer(3, ProtectedHighsBuffer, INDICATOR_DATA);
   SetIndexBuffer(4, ProtectedLowsBuffer, INDICATOR_DATA);
   SetIndexBuffer(5, LineToFlipBullish, INDICATOR_DATA);
   SetIndexBuffer(6, LineToFlipBearish, INDICATOR_DATA);
   SetIndexBuffer(7, LineForNewZone, INDICATOR_DATA);

   if(plotHighLow)
     {
      PlotIndexSetInteger(0, PLOT_DRAW_TYPE, DRAW_ARROW);
      PlotIndexSetInteger(0, PLOT_LINE_COLOR, clrBlack);
      PlotIndexSetInteger(0, PLOT_ARROW, 159);
      PlotIndexSetInteger(0, PLOT_ARROW_SHIFT, -15);
      PlotIndexSetString(0, PLOT_LABEL, "Highs");

      PlotIndexSetInteger(1, PLOT_DRAW_TYPE, DRAW_ARROW);
      PlotIndexSetInteger(1, PLOT_LINE_COLOR, clrBlack);
      PlotIndexSetInteger(1, PLOT_ARROW, 159);
      PlotIndexSetInteger(1, PLOT_ARROW_SHIFT, +15);
      PlotIndexSetString(1, PLOT_LABEL, "Lows");
     }
   else
     {
      PlotIndexSetInteger(0, PLOT_DRAW_TYPE, DRAW_NONE);
      PlotIndexSetInteger(1, PLOT_DRAW_TYPE, DRAW_NONE);
     }
   if(plotVHLValidation)
     {
      PlotIndexSetInteger(2, PLOT_DRAW_TYPE,DRAW_LINE);
      PlotIndexSetInteger(2, PLOT_LINE_COLOR, clrOrange);
      PlotIndexSetInteger(2, PLOT_LINE_WIDTH, 1);
      PlotIndexSetInteger(2,PLOT_LINE_STYLE,0);
      PlotIndexSetString(2, PLOT_LABEL, "Validation Level");
     }
   else
     {
      PlotIndexSetInteger(2, PLOT_DRAW_TYPE, DRAW_NONE);
     }
   if(plotProtectedHighlow)
     {
      PlotIndexSetInteger(3, PLOT_DRAW_TYPE, DRAW_ARROW); // Using arrows to represent protected highs
      PlotIndexSetInteger(3, PLOT_LINE_COLOR, clrGreen); // Red color to differentiate from regular highs
      PlotIndexSetInteger(3, PLOT_ARROW, 117); // Different arrow code, choose one appropriate from Wingdings
      PlotIndexSetInteger(3, PLOT_ARROW_SHIFT, -30); // Shift to align with your chart setup
      PlotIndexSetString(3, PLOT_LABEL, "Protected Highs"); // Label for the plot

      PlotIndexSetInteger(4, PLOT_DRAW_TYPE, DRAW_ARROW); // Using arrows to represent protected lows
      PlotIndexSetInteger(4, PLOT_LINE_COLOR, clrFireBrick); // Green color to differentiate from regular lows
      PlotIndexSetInteger(4, PLOT_ARROW, 117); // Different arrow code, choose one appropriate from Wingdings
      PlotIndexSetInteger(4, PLOT_ARROW_SHIFT, +30); // Shift to align with your chart setup
      PlotIndexSetString(4, PLOT_LABEL, "Protected Lows"); // Label for the plot
     }
   else
     {
      PlotIndexSetInteger(3, PLOT_DRAW_TYPE, DRAW_NONE);
      PlotIndexSetInteger(4, PLOT_DRAW_TYPE, DRAW_NONE);
     }

   if(plotcurrentrange)
     {
      PlotIndexSetInteger(5, PLOT_DRAW_TYPE,DRAW_LINE);
      PlotIndexSetInteger(5, PLOT_LINE_COLOR, clrGreen);
      PlotIndexSetInteger(5, PLOT_LINE_WIDTH, 1);
      PlotIndexSetString(5, PLOT_LABEL, "Range High");

      PlotIndexSetInteger(6, PLOT_DRAW_TYPE,DRAW_LINE);
      PlotIndexSetInteger(6, PLOT_LINE_COLOR, clrFireBrick);
      PlotIndexSetInteger(6, PLOT_LINE_WIDTH, 1);
      PlotIndexSetString(6, PLOT_LABEL, "Range Low");
     }
   else
     {
      PlotIndexSetInteger(5, PLOT_DRAW_TYPE, DRAW_NONE);
      PlotIndexSetInteger(6, PLOT_DRAW_TYPE, DRAW_NONE);
     }

   if(plotrangebreak)
     {
      PlotIndexSetInteger(7, PLOT_DRAW_TYPE,DRAW_LINE);
      PlotIndexSetInteger(7, PLOT_LINE_COLOR, clrBlack);
      PlotIndexSetInteger(7, PLOT_LINE_WIDTH, 1);
      PlotIndexSetInteger(7,PLOT_LINE_STYLE,2);
      PlotIndexSetString(7, PLOT_LABEL, "Level for a new zone");
     }
   else
     {
      PlotIndexSetInteger(7, PLOT_DRAW_TYPE, DRAW_NONE);
     }
   return INIT_SUCCEEDED;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void OnDeinit(const int reason)
  {
   ArrayResize(HighsBuffer, 0);
   ArrayResize(LowsBuffer, 0);
   ArrayResize(LineBuffer, 0);
   ArrayResize(LineToFlipBullish, 0);
   ArrayResize(LineToFlipBearish, 0);
   ArrayResize(ProtectedHighsBuffer, 0);
   ArrayResize(ProtectedLowsBuffer, 0);
   ObjectsDeleteAll(ChartID(), 0, OBJ_TREND);
   ObjectsDeleteAll(ChartID(), 0, OBJ_TEXT);
//Print("Indicator buffers have been cleared on deinitialization.");
  }
//+------------------------------------------------------------------+
//| Calculation function                                             |
//+------------------------------------------------------------------+
int OnCalculate(const int rates_total,
                const int prev_calculated,
                const datetime &time[],
                const double &open[],
                const double &high[],
                const double &low[],
                const double &close[],
                const long &tick_volume[],
                const long &volume[],
                const int &spread[])
  {
   int start = MathMax(rates_total - int(daysback * 24 * 60 / (PeriodSeconds(_Period) / 60)), 0);
   start = MathMax(start, prev_calculated - 2);

   if(rates_total > 0)
     {
      HighsBuffer[rates_total - 1] = EMPTY_VALUE;
      LowsBuffer[rates_total - 1] = EMPTY_VALUE;
      LineBuffer[rates_total - 1] = EMPTY_VALUE;
      ProtectedHighsBuffer[rates_total - 1]= EMPTY_VALUE;
      ProtectedLowsBuffer[rates_total - 1]= EMPTY_VALUE;
      LineToFlipBullish[rates_total - 1]= EMPTY_VALUE;
      LineToFlipBearish[rates_total - 1]= EMPTY_VALUE;
      LineForNewZone[rates_total - 1]= EMPTY_VALUE;
      //fibonacciHighs[rates_total - 1]= EMPTY_VALUE;
      //fibonacciLows[rates_total - 1]= EMPTY_VALUE;
     }

   for(int i = 0; i < rates_total- int(daysback * 24 * 60); i++)
     {
      HighsBuffer[i] = EMPTY_VALUE;
      LowsBuffer[i] = EMPTY_VALUE;
      LineBuffer[i] = EMPTY_VALUE;
      ProtectedHighsBuffer[i]= EMPTY_VALUE;
      ProtectedLowsBuffer[i]= EMPTY_VALUE;
      LineToFlipBullish[i]=EMPTY_VALUE;
      LineToFlipBearish[i]=EMPTY_VALUE;
      LineForNewZone[i]= EMPTY_VALUE;
      //fibonacciHighs[i]= EMPTY_VALUE;
      //fibonacciLows[i]= EMPTY_VALUE;
     }

//============================================
// logic to start searching for highs and lows
//============================================

   for(int i = start; i < rates_total-1; i++)
     {
      bool continueProcessing = false;
      do
        {
         continueProcessing = processBar(i, rates_total, high, low, close, open);
         continueProcessing = updateProtectedValues(i, rates_total, high, low, close, open,start);
        }
      while(continueProcessing);
      LineBuffer[i] = currentVC;//The line that will validate a high/low
      LineForNewZone[i]= currentLineForNewZone;//The line that when breached a new zone will be determined
      LineToFlipBullish[i] = currentLineToFlipBullish;//The line that when breached, structure will shift bullish
      LineToFlipBearish[i] = currentLineToFlipBearish ;//The line that when breached, structure will shift bearish
     }
   PlotFibonacciLevels(fibHighLevel,fibLowLevel,fibStartIndex,fibEndIndex);

   return rates_total;
  }

//+------------------------------------------------------------------+
//| FUNCTIONS                                                        |
//+------------------------------------------------------------------+
bool processBar(int i, int rates_total, const double &high[], const double &low[], const double &close[], const double &open[])
  {
   bool reevaluate = false; // Flag to determine if the bar should be re-evaluated

// Reset buffer values for the current bar to avoid drawing unwanted arrows
   HighsBuffer[i] = EMPTY_VALUE;
   LowsBuffer[i] = EMPTY_VALUE;

//------------------------------------------------
// Validation of a highs
//------------------------------------------------

   if(close[i] < VLH && VLH < 1E+18)
     {
      if(prevvalidHighindex != -1)
        {
         HighsBuffer[prevvalidHighindex] = EMPTY_VALUE;
        }
      HighsBuffer[validHighIndex] = high[validHighIndex];
      prevvalidHighindex = validHighIndex;
      prevvalidLowindex = -1;
      if(highbreached || (rangeValidHighIndex ==-1 && rangeValidLowIndex !=-1))
        {
         rangeValidHighIndex = validHighIndex;
         for(int k = rangeValidHighIndex; k < i; k++)
           {
            LineForNewZone[k]=HighsBuffer[rangeValidHighIndex];
           }
         LineForNewZone[rangeValidHighIndex-1]=EMPTY_VALUE;
         currentLineForNewZone = HighsBuffer[rangeValidHighIndex];
         highbreached=false;
        }
      VLH = 1E+18;

      lowestLow=1E+18;
      lowfound = false;
      reevaluate = true;
     }
//------------------------------------------------
// Validation of a lows
//------------------------------------------------
   if(close[i] > VLL && VLL > -1)
     {
      if(prevvalidLowindex != -1)
        {
         LowsBuffer[prevvalidLowindex] = EMPTY_VALUE;
        }
      LowsBuffer[validLowIndex] = low[validLowIndex];
      prevvalidLowindex = validLowIndex;
      prevvalidHighindex = -1;
      if(lowbreached || (rangeValidHighIndex ==-1 && rangeValidLowIndex ==-1))
        {
         rangeValidLowIndex = validLowIndex;
         for(int k = rangeValidLowIndex; k < i; k++)
           {
            LineForNewZone[k] = LowsBuffer[rangeValidLowIndex];

           }
         LineForNewZone[rangeValidLowIndex-1]=EMPTY_VALUE;
         currentLineForNewZone = LowsBuffer[rangeValidLowIndex];
         lowbreached=false;
        }
      VLL = -1;
      highestHigh =-1;
      lowfound = true;
      reevaluate = true;
     }

//------------------------------------------
//------------------------------------------
// Finding valid high logic
//------------------------------------------
//------------------------------------------
   if(close[i] > highestHigh+margin && close[i] > open[i])
     {
      highestHigh = high[i];
      ArrayInitialize(LineBuffer, EMPTY_VALUE);
      VLH = low[i];
      currentVC = VLH; // Update global VLH tracker
      validHighIndex = i;
      gapup_search = true;
      return false;
     }

   if(high[i] > highestHigh+margin && close[i]<highestHigh-margin && gapup_search && highestHigh>-1 && i == validHighIndex+1)
     {
      if(open[i] > close[i - 1]+margin)
        {
         highestHigh = high[i];
         validHighIndex = i;
         return false;
        }
      else
        {
         gapup_search=false;
         return false;
        }
     }

//------------------------------------------
//------------------------------------------
// Finding valid low logic
//------------------------------------------
//------------------------------------------
// If a valid low has been recorded and a new lower value is found
   if(close[i] < lowestLow-margin && close[i] < open[i])
     {
      lowestLow = low[i];
      ArrayInitialize(LineBuffer, EMPTY_VALUE);
      VLL = high[i];  // Update VLL as well
      currentVC = VLL; // Update global VLL tracker
      validLowIndex = i;
      gapdown_search = true;
      return false;
     }

// When another lower value is found and the previous close is higher than the lowest low
   if(low[i] < lowestLow-margin && close[i] > lowestLow+margin && gapdown_search && lowestLow < 1E+18 && i == validLowIndex+1)
     {
      if(open[i] < close[i - 1]-margin)
        {
         lowestLow = low[i];
         validLowIndex = i;
         return false;
        }
      else
        {
         gapdown_search = false;
         return false;
        }
     }
   return reevaluate; // Return the flag status
  }

//+------------------------------------------------------------------+
//|  BREAKOUT LOGIC THAT WILL PROVIDE PROTECTED HIGHS AND LOWS                                                                |
//+------------------------------------------------------------------+
bool updateProtectedValues(int i, const int rates_total, const double &high[], const double &low[], const double &close[], const double &open[], int start_local)
  {
   bool reevaluate = false; // Flag to determine if the bar should be re-evaluated

   ProtectedHighsBuffer[i] = EMPTY_VALUE;
   ProtectedLowsBuffer[i] = EMPTY_VALUE;

   if(!highbreached && rangeValidHighIndex !=-1 && rangeValidLowIndex!=-1 && close[i] > HighsBuffer[rangeValidHighIndex]+margin)
     {
      if(!lowfound)
        {
         HighsBuffer[prevvalidHighindex] = EMPTY_VALUE;
         VLH = 1E+18;
         highestHigh =-1;
        }
      double protectedLow = 1E+18;
      int protectedLowIndex = -1;
      breakoutHighIndex = i+1;

      for(int j = rangeValidHighIndex; j <= i; j++)
        {
         if(LowsBuffer[j] != EMPTY_VALUE && LowsBuffer[j] < protectedLow)
           {
            protectedLow = LowsBuffer[j];
            protectedLowIndex = j;
            //now you have located the index and the value of the protected low
           }
        }

      //this part will push that index & the value of the protected low found above into the buffers
      if(protectedLowIndex != -1)
        {
         ProtectedLowsBuffer[protectedLowIndex] = protectedLow;
         rangeValidLowIndex = protectedLowIndex;

         for(int k = protectedLowIndex; k < i; k++)
           {
            LineToFlipBearish[k]=protectedLow;//this is to draw the line range low backwards to the protected low
           }
         currentLineForNewZone = EMPTY_VALUE;//to make sure no skewd line is drawn
         LineToFlipBearish[protectedLowIndex-1]=EMPTY_VALUE;//to make sure no skewd line is drawn
         currentLineToFlipBearish=protectedLow;//This is the line that if breached will make structure bearish
         currentLineToFlipBullish = EMPTY_VALUE;//to make sure no skewd line is drawn
         fibStartIndex = rates_total-breakoutHighIndex;
         fibHighLevel = HighsBuffer[rangeValidHighIndex];
         fibLowLevel = ProtectedLowsBuffer[protectedLowIndex];
        }
      else
        {
         for(int l = rangeValidHighIndex; l <= i; l++)
           {
            LineForNewZone[l]=EMPTY_VALUE;
           }
        }
      highbreached = true;
      return true;
     }

   if(!lowbreached && rangeValidLowIndex!=-1 && rangeValidHighIndex !=-1 && close[i] < LowsBuffer[rangeValidLowIndex]-margin)
     {
      if(lowfound)
        {
         LowsBuffer[prevvalidLowindex] = EMPTY_VALUE;
         VLL=-1;
         lowestLow=1E+18;
        }
      double protectedHigh = -1;
      int protectedHighIndex = -1;
      breakoutLowIndex = i-1;

      for(int j = rangeValidLowIndex; j <= i; j++)
        {
         if(HighsBuffer[j] != EMPTY_VALUE && HighsBuffer[j] > protectedHigh)
           {
            protectedHigh = HighsBuffer[j];
            protectedHighIndex = j;
           }
        }
      if(protectedHighIndex != -1)
        {
         ProtectedHighsBuffer[protectedHighIndex] = protectedHigh;
         rangeValidHighIndex = protectedHighIndex;
         //Print("Protected High updated at index: ", protectedHighIndex, " with value: ", protectedHigh);

         for(int k = protectedHighIndex; k < i; k++)
           {
            LineToFlipBullish[k]=protectedHigh;
           }
         currentLineForNewZone = EMPTY_VALUE;//to make sure no skewd line is drawn
         LineToFlipBullish[protectedHighIndex-1]=EMPTY_VALUE;//to make sure no skewd line is drawn
         currentLineToFlipBullish=protectedHigh;//This is the line that if breached will make structure bullish
         currentLineToFlipBearish = EMPTY_VALUE;//to make sure no skewd line is drawn
         fibStartIndex = rates_total-breakoutLowIndex;
         fibHighLevel = ProtectedHighsBuffer[protectedHighIndex];
         fibLowLevel = LowsBuffer[rangeValidLowIndex];
        }
      else
        {
         for(int l = rangeValidLowIndex; l <= i; l++)
           {
            LineForNewZone[l]=EMPTY_VALUE;
           }
        }
      lowbreached = true;
      return true;
     }




   return reevaluate;
  }

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void PlotFibonacciLevels(double high, double low, int startIndex, int endIndex)
  {
   if(startIndex < 0 || startIndex >= Bars(_Symbol, _Period))
     {
      Print("Invalid start index provided.");
      return;
     }

   double range = NormalizeDouble(high - low, 2);

// Clear existing Fibonacci lines first
   ObjectsDeleteAll(ChartID(), 0, OBJ_TREND);
   ObjectsDeleteAll(ChartID(), 0, OBJ_TEXT);

// Check each level and create a trend line if enabled
   if(useLevel1)
      CreateFibTrendLine("FibLevel1", NormalizeDouble(high - range * (level1/100), 2), startIndex, endIndex, clrBlack, DoubleToString(level1, 0) + "%");
   if(useLevel2)
      CreateFibTrendLine("FibLevel2", NormalizeDouble(high - range * (level2/100), 2), startIndex, endIndex, clrBlack, DoubleToString(level2, 0) + "%");
   if(useLevel3)
      CreateFibTrendLine("FibLevel3", NormalizeDouble(high - range * (level3/100), 2), startIndex, endIndex, clrBlack, DoubleToString(level3, 0) + "%");

  }

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void CreateFibTrendLine(string name, double level, int startIndex, int endIndex, color lineColor, string labelText)
  {
   datetime startTime = iTime(_Symbol, _Period, startIndex);
   datetime endTime = iTime(_Symbol, _Period, endIndex);
   datetime endTimeText = endTime + PeriodSeconds(_Period) * 5;

   if(!ObjectCreate(0, name, OBJ_TREND, 0, startTime, level, endTime, level))
     {
      Print("Failed to create Fibonacci trend line: ", name);
      return;
     }
   ObjectSetInteger(0, name, OBJPROP_COLOR, lineColor);
   ObjectSetInteger(0, name, OBJPROP_WIDTH, 1);
   ObjectSetInteger(0, name, OBJPROP_STYLE, STYLE_DASHDOT);

// Create a text label at the end of the line
   string textName = name + "Text";
   if(!ObjectCreate(0, textName, OBJ_TEXT, 0, endTimeText, level))
     {
      Print("Failed to create text label for Fibonacci trend line: ", textName);
      return;
     }
   ObjectSetString(0, textName, OBJPROP_TEXT, labelText);
   ObjectSetInteger(0, textName, OBJPROP_COLOR, lineColor);
   ObjectSetInteger(0, textName, OBJPROP_FONTSIZE, 8);
   ObjectSetInteger(0, textName, OBJPROP_ANCHOR, ANCHOR_RIGHT);
  }
//+------------------------------------------------------------------+

//+------------------------------------------------------------------+
